﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using Ozeki.Media;
using Ozeki.Media.Codec;
using Ozeki.Media.MediaHandlers;
using Ozeki.Media.MediaHandlers.Video;
using Ozeki.Media.Video;
using Ozeki.Network.Nat;
using Ozeki.VoIP;
using Ozeki.VoIP.SDK;

namespace VideoSoftphone
{
    public partial class VideoSoftphone : Form
    {

        private ISoftPhone softPhone;
        private PhoneCallAudioReceiver phoneCallAudioReceiver = new PhoneCallAudioReceiver();
        private PhoneCallAudioSender phoneCallAudioSender = new PhoneCallAudioSender();
        private PhoneCallVideoSender phonecallVideoSender = new PhoneCallVideoSender();
        private PhoneCallVideoReceiver phonecallVideoReceiver = new PhoneCallVideoReceiver();
        private MediaConnector mediaConnector = new MediaConnector();

        private Microphone microphone = Microphone.GetDefaultDevice();
        private Speaker speaker = Speaker.GetDefaultDevice();

        private List<VideoDeviceInfo> cameras = WebCamera.GetDevices();
        private IEnumerable<CodecInfo> codecs;

        private WebCamera camera = WebCamera.GetDefaultDevice();

        private ImageProvider<Image> remoteImageHandler = new DrawingImageProvider();
        private ImageProvider<Image> localImageHandler = new DrawingImageProvider();

        IPhoneCall call;
        bool inComingCall;
        PhoneLineState phoneLineInformation;
        IPhoneLine phoneLine;

        AdvancedSettings adv = new AdvancedSettings();

        public VideoSoftphone()
        {
            InitializeComponent();
            //some devices are missing
            string message = String.Empty;

            if (microphone == null)
                message += "You have no microphone attached to your computer, please note that your partner will not hear your voice.\n";
            if (speaker == null)
                message += "You have no speaker attached to your computer, please note that you will not hear your partner.\n";
            if (camera == null)
                message += "You have no camera attached to your computer, please note that you will not hear your partner.\n";

            if (message != String.Empty)
                MessageBox.Show(message, "Some devices are missing", MessageBoxButtons.OK, MessageBoxIcon.Information);
            setCameras();
            UpdateCameraResolution(comboBox1.SelectedItem as VideoDeviceInfo);
        }

        private void setCameras()
        {
            foreach (Ozeki.Media.Video.VideoDeviceInfo cam in cameras)
                this.comboBox1.Items.Add(cam);
            if (cameras.Count > 0)
                this.comboBox1.SelectedIndex = 0;
            camera = WebCamera.GetDevice(comboBox1.SelectedItem as VideoDeviceInfo);
        }

        private void UpdateCameraResolution(VideoDeviceInfo device)
        {
            if (device == null)
                return;
            comboBox2.Items.Clear();
            if (device.Capabilities != null)
            {
                foreach (VideoCapabilities resolution in device.Capabilities)
                {
                    comboBox2.Items.Add(resolution);
                }
                if (comboBox2.Items.Count > 0)
                    comboBox2.SelectedItem = device.Capabilities[0];

                comboBox2.Update();
            }
        }

        private void setCodecs()
        {
            if (softPhone == null)
                return;
            codecs = softPhone.Codecs;
            foreach (CodecInfo ci in codecs)
                comboBox3.Items.Add(ci);
            if (comboBox3.Items.Count > 0)
                comboBox3.SelectedIndex = 0;
        }

        #region Local camera testing

        private bool localVideoTest = false;
        
        private void btnVideoLocalTest_Click(object sender, EventArgs e)
        {
            localVideoTest = !localVideoTest;
            
            if (localVideoTest)
            {   
                if (camera != null)
                camera.Start();
                btnVideoLocalTest.Text = "Stop test";
                mediaConnector.Connect(camera, localImageHandler);
                videoViewerLocal.Start();
            }
            else
            {
                //VideoDevices.Text = "Device: ";
                
                btnVideoLocalTest.Text = "Local test";
                if (camera != null)
                camera.Stop();
                mediaConnector.Disconnect(camera, localImageHandler);
                videoViewerLocal.Stop();
            }
        }

    #endregion

        private void InvokeGUIThread(Action action)
        {
            Invoke(action);
        }


        private void InitializeSoftPhone()
        {
            try
            {
              
                softPhone = SoftPhoneFactory.CreateSoftPhone(SoftPhoneFactory.GetLocalIP(), 5700, 5750, 5700);
                softPhone.IncomingCall += new EventHandler<VoIPEventArgs<IPhoneCall>>(softPhone_IncomingCall);
                setCodecs();
                
                SIPAccount sa = new SIPAccount(adv.IsRegRequired, adv.displayname, adv.username, adv.registername, adv.regpass, adv.domainhost, adv.port);
                NatConfiguration nc = new NatConfiguration(adv.natTraversal);
                phoneLine = softPhone.CreatePhoneLine(sa, nc);

                phoneLine.PhoneLineStateChanged += new EventHandler<VoIPEventArgs<PhoneLineState>>(phoneLine_PhoneLineInformation);
                
                softPhone.RegisterPhoneLine(phoneLine);
                DialedNumber.Text = string.Empty;

                videoViewerLocal.SetImageProvider(localImageHandler);
                videoViewerRemote.SetImageProvider(remoteImageHandler);
            }
            catch (Exception ex)
            {
                MessageBox.Show(String.Format("You didn't give your local IP adress. Please press the Register button and try it again.\n {0}", ex.Message), string.Empty, MessageBoxButtons.OK,
                MessageBoxIcon.Error);
            }
        }


        #region Events&Handling


        private void phoneLine_PhoneLineInformation(object sender, VoIPEventArgs<PhoneLineState> e)
        {
            phoneLineInformation = e.Item;
            InvokeGUIThread(() =>
            {
                Username.Text = ((IPhoneLine)sender).SIPAccount.RegisterName;
                if (e.Item == PhoneLineState.RegistrationSucceeded)
                {
                    ConnectionState.Text = "Online";
                    RegistrationState.Text = "Registration succeeded";
                }
                else
                    RegistrationState.Text = e.Item.ToString();
            });
        }

        private void softPhone_IncomingCall(object sender, VoIPEventArgs<IPhoneCall> e)
        {
            InvokeGUIThread(() =>
            {
                RegistrationState.Text = "Incoming call";
                DialedNumber.Text = String.Format("from {0}", e.Item.DialInfo);
                call = e.Item;
                WireUpCallEvents();
                inComingCall = true;
            });
        }

        private void WireUpCallEvents()
        {
            call.CallStateChanged += (call_CallStateChanged);
            call.CallErrorOccured += (call_CallErrorOccured);
        }

        private void WireDownCallEvents()
        {
            if (call != null)
            {
                call.CallStateChanged -= (call_CallStateChanged); ;
                call.CallErrorOccured -= (call_CallErrorOccured);
            }
        }

        private void call_CallStateChanged(object sender, VoIPEventArgs<CallState> e)
        {
            InvokeGUIThread(() => { RegistrationState.Text = e.Item.ToString(); });

            switch (e.Item)
            {
                case CallState.InCall:
                                        
                    phoneCallAudioSender.AttachToCall(call);
                    phoneCallAudioReceiver.AttachToCall(call);
                    phonecallVideoSender.AttachToCall(call);
                    phonecallVideoReceiver.AttachToCall(call);
                    if (microphone != null)
                        microphone.Start();
                    mediaConnector.Connect(microphone, phoneCallAudioSender);
                    if (speaker != null)
                        speaker.Start();
                    mediaConnector.Connect(phoneCallAudioReceiver, speaker);
                    if (camera != null)
                        camera.Start();
                    mediaConnector.Connect(camera, localImageHandler);
                    mediaConnector.Connect(camera, phonecallVideoSender);
                    mediaConnector.Connect(phonecallVideoReceiver, remoteImageHandler);
                    

                    videoViewerLocal.Start();
                    videoViewerRemote.Start();
                    break;
                case CallState.Completed:
                    if (microphone != null)
                        microphone.Stop();
                    mediaConnector.Disconnect(microphone, phoneCallAudioSender);
                    if (speaker != null)
                    speaker.Stop();
                    mediaConnector.Disconnect(phoneCallAudioReceiver, speaker);
                    if (camera != null)
                    camera.Stop();
                    mediaConnector.Disconnect(camera, localImageHandler);
                    mediaConnector.Disconnect(camera, phonecallVideoSender);
                    mediaConnector.Disconnect(phonecallVideoReceiver, remoteImageHandler);

                    phoneCallAudioSender.Detach();
                    phoneCallAudioReceiver.Detach();
                    phonecallVideoSender.Detach();
                    phonecallVideoReceiver.Detach();

                    WireDownCallEvents();
                    call = null;

                    InvokeGUIThread(() => { DialedNumber.Text = string.Empty; });
                    break;
                case CallState.Cancelled:
                    WireDownCallEvents();
                    call = null;
                    break;
            }
        }

        private void call_CallErrorOccured(object sender, VoIPEventArgs<CallError> e)
        {
            InvokeGUIThread(() =>
            {
                RegistrationState.Text = e.Item.ToString();
            });
        }

        private void Softphone_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (call != null)
                HangUp.PerformClick();
        }


        #endregion

        #region KeyPadEvents

        private void PickUp_Click(object sender, EventArgs e)
        {
            if (inComingCall)
            {
                inComingCall = false;
                call.Accept();
                return;
            }

            if (call != null)
                return;

            if (string.IsNullOrEmpty(DialedNumber.Text))
                return;

            if (phoneLineInformation != PhoneLineState.RegistrationSucceeded && phoneLineInformation != PhoneLineState.NoRegNeeded)
            {
                MessageBox.Show("Phone line state is not valid!");
                return;
            }

            call = softPhone.CreateCallObject(phoneLine, DialedNumber.Text);
            WireUpCallEvents();
            call.Start();
        }

        private void HangUp_Click(object sender, EventArgs e)
        {
            if (call != null)
            {
                if (inComingCall && call.CallState == CallState.Ringing)
                    call.Reject();
                else
                    call.HangUp();
                inComingCall = false;
                call = null;
            }
            DialedNumber.Text = string.Empty;
        }

        private void KeyPadButton_Click(object sender, EventArgs e)
        {
            var btn = sender as Button;
            if (btn != null)
            {
                if (btn.Text == "0/+")
                {
                    DialedNumber.Text += "0";
                }
                else
                    DialedNumber.Text += btn.Text.Trim();
            }
        }

        #endregion

        private void VideoSoftphone_Load(object sender, EventArgs e)
        {
            adv.ShowDialog();
            InitializeSoftPhone();
        }

        private void Register_Click(object sender, EventArgs e)
        {
            adv.ShowDialog();
            InitializeSoftPhone();
        }

        private void btnStartVideo_Click(object sender, EventArgs e)
        {
            if (call == null)
                return;
            if (camera != null)
            camera.Start();
            mediaConnector.Connect(camera, localImageHandler);
            mediaConnector.Connect(camera, phonecallVideoSender);
            mediaConnector.Connect(phonecallVideoReceiver, remoteImageHandler);


            videoViewerLocal.Start();
            btnStartVideo.Enabled = false;
            btnStopVideo.Enabled = true;
            btnVideoLocalTest.Enabled = false;
        }

        private void btnStopVideo_Click(object sender, EventArgs e)
        {
            if (camera != null)
                camera.Stop();
            videoViewerLocal.Stop();
            mediaConnector.Disconnect(camera, localImageHandler);
            mediaConnector.Disconnect(camera, phonecallVideoSender);
            mediaConnector.Disconnect(phonecallVideoReceiver, remoteImageHandler);
            localImageHandler.Dispose();
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            camera = WebCamera.GetDevice(comboBox1.SelectedItem as VideoDeviceInfo);
            UpdateCameraResolution(comboBox1.SelectedItem as VideoDeviceInfo);
        }

        private void comboBox2_SelectedIndexChanged(object sender, EventArgs e)
        {
            VideoCapabilities vid = comboBox2.SelectedItem as VideoCapabilities;
            if (vid != null)
                camera.Resolution = vid.Resolution;
        }

        private void comboBox3_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (softPhone == null)
                return;
            CodecInfo ci = comboBox3.SelectedItem as CodecInfo;
            softPhone.EnableCodec(ci.Payload);
        }

    }
}
